/* cruelty :) */

/* Copyright (C) 2002 Grigory Bakunov <black@asplinux.ru> */

/* Copyright (C) 1997 Ian Main
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gtk/gtk.h>

#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "list.h"
#include "interface.h"
#include "dillo.h"
#include "nav.h"
#include "browser.h"
#include "menu.h"
#include "misc.h"
#include "url.h"

#define LOAD_ENCODINGS 1
#define SAVE_ENCODING 2
#define CLOSE_ENCODINGS 3

/* this #define will cut page title if > 39 chars */
#define TITLE39

/* double quote */
#define D_QUOTE 0x22

/* Data types */

typedef struct _Encoding Encoding;
typedef struct _CallbackInfo CallbackInfo;

struct _Encoding {
   char *title;
   char *iconv_name;
   GtkWidget *menuitem;
};

struct _CallbackInfo {
   BrowserWindow *bw;
   guint index;
};

/*
 * Forward declarations
 */
static void
   Encodings_load_to_menu(FILE *fp),
   Encodings_file_op(gint operation, const char *title, const char *iconv_name),
   Encodings_save_to_file(FILE *fp, const char *title,  const char *iconv_name);

static char*
   Encodings_search_line(char *line, char *start_text, char *end_text);

extern BrowserWindow **browser_window;

/*
 * Local data
 */
static Encoding *encodings = NULL;
static gint num_encodings = 0;
static gint num_encodings_max = 16;


/*
 * Allocate memory and load the encodings list
 */
void a_Encodings_init(void)
{
   gchar *file;

   /* Here we load and set the encodings */
   file = a_Misc_prepend_user_home(".dillo/encodings");
   if (access(file, R_OK))
	Encodings_file_op(LOAD_ENCODINGS, "/etc/dillo/encodings", NULL);
   else
   Encodings_file_op(LOAD_ENCODINGS, file, NULL);
   g_free(file);
}

/*
 * ?
 */
void Encodings_set_encoding(GtkWidget *widget, CallbackInfo *CbInfo)
{
   if (CbInfo->index >= num_encodings) {
      g_warning("encoding not found!\n");
      return;
   }
   g_print("Selected encoding: %s\n",encodings[CbInfo->index].iconv_name);
   CbInfo->bw->encoding=encodings[CbInfo->index].iconv_name;
   a_Nav_reload(CbInfo->bw);
   /*HERE!!! LOOKOUT!!!! */
}

/*
 * Add a encoding to the encodings menu of a particular browser window
 */
void Encodings_add_to_menu(BrowserWindow *bw, GtkWidget *menuitem, guint index)
{
   CallbackInfo *CbInfo;

   gtk_menu_append(GTK_MENU(bw->enc_menu), menuitem);

   CbInfo = g_new(CallbackInfo, 1);
   CbInfo->bw = bw;
   CbInfo->index = index;

   /* accelerator goes here */
   gtk_signal_connect(GTK_OBJECT (menuitem), "activate",
                      (GtkSignalFunc)Encodings_set_encoding, CbInfo);
}

/*
 * ?
 */
static GtkWidget *Encodings_insert(const char *title, const char *iconv_name)
{
   GtkWidget *menuitem;

   menuitem = gtk_menu_item_new_with_label(title);
   gtk_widget_show(menuitem);

   a_List_add(encodings, num_encodings, num_encodings_max);
   encodings[num_encodings].title = g_strdup(title);
   encodings[num_encodings].iconv_name = g_strdup(iconv_name);
   encodings[num_encodings].menuitem = menuitem;
   num_encodings++;
   return menuitem;
}

/*
 * Add the new encoding to encodings menu of _all_ browser windows and then
 * write the new encoding to file
 */
/*
void a_Encodings_add(GtkWidget *widget, gpointer client_data)
{
   BrowserWindow *bw = (BrowserWindow *)client_data;
   gint i;
#ifdef TITLE39
   gboolean allocated = FALSE;
#endif
   char *title;
   char *iconv_name;
   GtkWidget *menuitem;

   title = bw->menu_popup.info.title;
   iconv_name = bw->menu_popup.info.iconv_name;

#ifdef TITLE39
   if (strlen (title) > 39) {
      char buf1[20];
      char buf2[20];

      memcpy (buf1, title, 18);
      buf1[18] = '\0';
      strcpy (buf2, title + strlen (title) - 18);
      buf2[18] = '\0';
      title = g_strconcat (buf1, "...", buf2, NULL);
      allocated = TRUE;
   }
#endif

   menuitem = Encodings_insert(title, iconv_name);
   Encodings_add_to_menu(browser_window[0], menuitem, num_encodings-1);
   for (i = 1; i < num_bw; i++) {
      menuitem= gtk_menu_item_new_with_label(encodings[num_encodings-1].title);
      gtk_widget_show(menuitem);
      Encodings_add_to_menu(browser_window[i], menuitem, num_encodings-1);
   }

   Encodings_file_op(SAVE_ENCODING, title, iconv_name);

#ifdef TITLE39
   if (allocated)
      g_free (title);
#endif
}
*/
/*
 * Never called (the file remains open all the time)  --Jcid
 */
void Encodings_close(void)
{
   Encodings_file_op(CLOSE_ENCODINGS, NULL, NULL);
}

/*
 * Performs operations on the encoding file..
 * for first call, title is the filename
 */
static void
 Encodings_file_op(gint operation, const char *title, const char *iconv_name)
{
   static FILE *fp;
   static gint initialized = 0;

   if (!initialized) {
      if (operation == LOAD_ENCODINGS) {
         /* SAVE_ENCODINGS is broken anyway, better make use of /etc/dillo/encodings */
         if ((fp = fopen(title, "r")) == NULL)
            g_print("dillo: opening encoding file %s: %s\n",
                    title, strerror(errno));
         else
            initialized = 1;
      } else
         g_print("Error: invalid call to Encodings_file_op.\n");
   }

   g_return_if_fail( initialized );

   switch (operation) {
   case LOAD_ENCODINGS:
      Encodings_load_to_menu(fp);
      break;

   case SAVE_ENCODING:
      Encodings_save_to_file(fp, title, iconv_name);
      break;

   case CLOSE_ENCODINGS:
      fclose(fp);
      break;

   default:
      break;
   }
}

/*
 * Save encodings to ~/.dillo/encodings
 */
static void
 Encodings_save_to_file(FILE *fp, const char *title, const char* iconv_name)
{
   fseek(fp, 0L, SEEK_END);
   fprintf(fp, "<enc value=\"%s\">%s</a>\n", iconv_name,title);
   fflush(fp);
}

/*
 * Load encodings
 */
static void Encodings_load_to_menu(FILE *fp)
{
   gchar *title=NULL;
   gchar *iconv_name=NULL;
   char buf[4096];
   gint i = 0;
   GtkWidget *menuitem;

   rewind(fp);

   g_print("Loading encodings...\n");
   while (1) {
      /* Read a whole line from the file */
      if ((fgets(buf, 4096, fp)) == NULL)
          break;

      /* get url from line */
      if ( !(iconv_name = Encodings_search_line(buf, "=\"", "\">")) )
         continue;

      /* get title from line */
      if ( !(title = Encodings_search_line(buf, "\">", "</")) ){
         g_free(iconv_name);
         continue;
      }
      printf("%s = %s\n",title,iconv_name);
      menuitem = Encodings_insert(title, iconv_name);
      Encodings_add_to_menu(browser_window[0], menuitem, i);
      g_free(iconv_name);
      g_free(title);
      i++;
   }
}

/*
 * Copy encodings when new browser windows are opened.
 * Called by 'a_Menu_mainbar_new()'
 */
void a_Encodings_fill_new_menu(BrowserWindow *bw)
{
   gint i;
   GtkWidget *menuitem;

   for (i = 0; i < num_encodings; i++) {
     menuitem = gtk_menu_item_new_with_label(encodings[i].title);
     gtk_widget_show(menuitem);
     Encodings_add_to_menu(bw, menuitem, i);
   }
}

static char *
 Encodings_search_line(char *line, char *start_text, char *end_text) {
   gint segment_length;
   char *start_index, *end_index;

   /* if string is not found, return NULL */
   if ((start_index = strstr(line, start_text)) == NULL)
      return (NULL);
   if ((end_index = strstr(line, end_text)) == NULL)
      return (NULL);

   /* adjustment cause strstr returns the start of the text */
   start_index += strlen(start_text);

   /* find length of text segment */
   segment_length = end_index - start_index;
   return g_strndup(start_index, segment_length);
};

void   a_Encoding_get_conversion(BrowserWindow *bw, char* charset) {
    bw->dcv.c_from=iconv_open(curr_locale,charset);
    bw->dcv.c_to=iconv_open(charset,curr_locale);
    if (bw->dcv.c_from == (iconv_t)-1 || bw->dcv.c_to == (iconv_t)-1)
	g_warning ("could not allocate character encoding converter.");
    return;
};

#define MAX_CHAR_SIZE 4
gchar *a_Encoding_translate_encoding(deConversion dcv, gchar *buf, gint bufsize,deDirection dir) {
    char *result, *source, *dest; 
    size_t s_left, d_left;
    iconv_t conversion;
    if (dir==DE_DECODE)
	conversion=dcv.c_from;
    else 
	conversion=dcv.c_to;
    /* no conversion is needed, or none is available */
    if (conversion == (iconv_t)-1)
	return g_strndup(buf, bufsize);

    /* Note that for some conversions, the translated buffer can be larger
     * than the input buffer. This is particularly important for conversions
     * to UTF8 (check the unicode standard to find out the scale factor). */
    result = g_malloc((bufsize + 1)*MAX_CHAR_SIZE);
    
    source = buf;
    dest = result;
    s_left =  bufsize;
    d_left = bufsize*MAX_CHAR_SIZE;
    if (iconv(conversion, &source, &s_left, &dest, &d_left)
	== (size_t)-1) {
	g_warning ("unable to fully convert page to native character set");
	/* This just skips past unconvertable characters, putting "?" in the
	 * output, then retries the conversion. This is a hack, but it seems
	 * like the best course of action in the circumstances. */
	while (s_left > 0 && d_left > 0 && errno == EILSEQ) {
	    source++;
	    s_left--;
	    *dest = '?';
	    dest++;
	    d_left--;
	    if (s_left > 0 && d_left > 0)
		iconv(conversion, &source, &s_left, &dest, &d_left);
	}
    }
    *dest = 0; /* terminate the string */
    return result;
}
